#include "common.h"


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <CL/cl.h>

#define BONES_MIN(a,b) ((a<b) ? a : b)
#define BONES_MAX(a,b) ((a>b) ? a : b)
#define DIV_CEIL(a,b)  ((a+b-1)/b)
#define DIV_FLOOR(a,b) (a/b)

// Load the OpenCL kernel from file
char * get_source(const char* filename) {
  FILE* fp = fopen(filename,"r");
  fseek(fp,0,SEEK_END);
  long size = ftell(fp);
  rewind(fp);
  char *source = (char *)malloc(sizeof(char)*(size+1));
  int temp = fread(source,1,sizeof(char)*size,fp);
  source[size] = '\0';
  fclose(fp);
  return source;
}

// Print an error if it occurs
void error_check(cl_int errors) {
  if(errors != CL_SUCCESS) {
    switch (errors) {
      case CL_DEVICE_NOT_FOUND:                 printf("--- Error: Device not found.\n"); break;
      case CL_DEVICE_NOT_AVAILABLE:             printf("--- Error: Device not available\n"); break;
      case CL_COMPILER_NOT_AVAILABLE:           printf("--- Error: Compiler not available\n"); break;
      case CL_MEM_OBJECT_ALLOCATION_FAILURE:    printf("--- Error: Memory object allocation failure\n"); break;
      case CL_OUT_OF_RESOURCES:                 printf("--- Error: Out of resources\n"); break;
      case CL_OUT_OF_HOST_MEMORY:               printf("--- Error: Out of host memory\n"); break;
      case CL_PROFILING_INFO_NOT_AVAILABLE:     printf("--- Error: Profiling information not available\n"); break;
      case CL_MEM_COPY_OVERLAP:                 printf("--- Error: Memory copy overlap\n"); break;
      case CL_IMAGE_FORMAT_MISMATCH:            printf("--- Error: Image format mismatch\n"); break;
      case CL_IMAGE_FORMAT_NOT_SUPPORTED:       printf("--- Error: Image format not supported\n"); break;
      case CL_BUILD_PROGRAM_FAILURE:            printf("--- Error: Program build failure\n"); break;
      case CL_MAP_FAILURE:                      printf("--- Error: Map failure\n"); break;
      case CL_INVALID_VALUE:                    printf("--- Error: Invalid value\n"); break;
      case CL_INVALID_DEVICE_TYPE:              printf("--- Error: Invalid device type\n"); break;
      case CL_INVALID_PLATFORM:                 printf("--- Error: Invalid platform\n"); break;
      case CL_INVALID_DEVICE:                   printf("--- Error: Invalid device\n"); break;
      case CL_INVALID_CONTEXT:                  printf("--- Error: Invalid context\n"); break;
      case CL_INVALID_QUEUE_PROPERTIES:         printf("--- Error: Invalid queue properties\n"); break;
      case CL_INVALID_COMMAND_QUEUE:            printf("--- Error: Invalid command queue\n"); break;
      case CL_INVALID_HOST_PTR:                 printf("--- Error: Invalid host pointer\n"); break;
      case CL_INVALID_MEM_OBJECT:               printf("--- Error: Invalid memory object\n"); break;
      case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:  printf("--- Error: Invalid image format descriptor\n"); break;
      case CL_INVALID_IMAGE_SIZE:               printf("--- Error: Invalid image size\n"); break;
      case CL_INVALID_SAMPLER:                  printf("--- Error: Invalid sampler\n"); break;
      case CL_INVALID_BINARY:                   printf("--- Error: Invalid binary\n"); break;
      case CL_INVALID_BUILD_OPTIONS:            printf("--- Error: Invalid build options\n"); break;
      case CL_INVALID_PROGRAM:                  printf("--- Error: Invalid program\n"); break;
      case CL_INVALID_PROGRAM_EXECUTABLE:       printf("--- Error: Invalid program executable\n"); break;
      case CL_INVALID_KERNEL_NAME:              printf("--- Error: Invalid kernel name\n"); break;
      case CL_INVALID_KERNEL_DEFINITION:        printf("--- Error: Invalid kernel definition\n"); break;
      case CL_INVALID_KERNEL:                   printf("--- Error: Invalid kernel\n"); break;
      case CL_INVALID_ARG_INDEX:                printf("--- Error: Invalid argument index\n"); break;
      case CL_INVALID_ARG_VALUE:                printf("--- Error: Invalid argument value\n"); break;
      case CL_INVALID_ARG_SIZE:                 printf("--- Error: Invalid argument size\n"); break;
      case CL_INVALID_KERNEL_ARGS:              printf("--- Error: Invalid kernel arguments\n"); break;
      case CL_INVALID_WORK_DIMENSION:           printf("--- Error: Invalid work dimensionsension\n"); break;
      case CL_INVALID_WORK_GROUP_SIZE:          printf("--- Error: Invalid work group size\n"); break;
      case CL_INVALID_WORK_ITEM_SIZE:           printf("--- Error: Invalid work item size\n"); break;
      case CL_INVALID_GLOBAL_OFFSET:            printf("--- Error: Invalid global offset\n"); break;
      case CL_INVALID_EVENT_WAIT_LIST:          printf("--- Error: Invalid event wait list\n"); break;
      case CL_INVALID_EVENT:                    printf("--- Error: Invalid event\n"); break;
      case CL_INVALID_OPERATION:                printf("--- Error: Invalid operation\n"); break;
      case CL_INVALID_GL_OBJECT:                printf("--- Error: Invalid OpenGL object\n"); break;
      case CL_INVALID_BUFFER_SIZE:              printf("--- Error: Invalid buffer size\n"); break;
      case CL_INVALID_MIP_LEVEL:                printf("--- Error: Invalid mip-map level\n"); break;
      default:                                  printf("--- Error: Unknown with code %d\n", errors);
    }
    fflush(stdout); exit(0);
  }
}

void pfn_notify(const char *errinfo, const void *private_info, size_t cb, void *user_data)
{
   printf("%s\n", errinfo);
}

// Function to initialize the OpenCL platform (create to ensure fair measurements afterwards)
void initialize_target(void) {
  cl_int errors;
  
  // Get OpenCL platform count
  cl_uint num_platforms;
  errors = clGetPlatformIDs(0,NULL,&num_platforms); error_check(errors);
  if (num_platforms == 0) { printf("Error: No OpenCL platforms found.\n"); exit(1); }
  
  // Get all OpenCL platform IDs
  cl_platform_id platform_ids[10];
  errors = clGetPlatformIDs(num_platforms,platform_ids,NULL); error_check(errors);  
  
  // Select the AMD APP platform
  char buffer[1024];
  cl_uint platform;
  for(cl_uint platform_id=0; platform_id<num_platforms; platform_id++) {
    clGetPlatformInfo(platform_ids[platform_id], CL_PLATFORM_NAME, 1024, buffer, NULL);
    if(strstr(buffer,"Intel") != NULL) { platform = platform_id; break; }
  }
  
  // Get a CPU device on the platform
  cl_device_id device;
  errors = clGetDeviceIDs(platform_ids[platform], CL_DEVICE_TYPE_CPU, 1, &device, NULL); error_check(errors);
  errors = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(buffer), buffer, NULL); error_check(errors);
  
  // Create a context
  cl_context_properties prop[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_ids[platform], 0 };
  printf("Creating an OpenCL context...\n"); fflush(stdout);
  cl_context context = clCreateContext(prop,1,&device,&pfn_notify,NULL,&errors); error_check(errors);
  //cl_context context = clCreateContextFromType(prop,CL_DEVICE_TYPE_CPU,&pfn_notify,NULL,&errors); error_check(errors);
  printf("...done!\n"); fflush(stdout);
  
  // Create space on the device
  cl_mem device_data = clCreateBuffer(context,CL_MEM_READ_WRITE,4,NULL,&errors); error_check(errors);
  
  // Copy something to the device
  device_data = clCreateBuffer(context,CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR,4,buffer,NULL);
  
  // Clean-up the OpenCL context
  clReleaseMemObject(device_data);
  clReleaseContext(context);
  fflush(stdout);
}

void saxpy_43_accelerated(int i, float a, float y[LARGE_N], float x[LARGE_N]) {
  fflush(stdout);
  cl_int errors;
  cl_event event;
  
  // Get OpenCL platform count
  cl_uint num_platforms;
  errors = clGetPlatformIDs(0,NULL,&num_platforms); error_check(errors);
  if (num_platforms == 0) { printf("Error: No OpenCL platforms found.\n"); exit(1); }
  
  // Get all OpenCL platform IDs
  cl_platform_id platform_ids[10];
  errors = clGetPlatformIDs(num_platforms,platform_ids,NULL); error_check(errors);  
  
  // Select the AMD APP platform
  char buffer[1024];
  cl_uint platform;
  for(cl_uint platform_id=0; platform_id<num_platforms; platform_id++) {
    clGetPlatformInfo(platform_ids[platform_id],CL_PLATFORM_NAME,1024,buffer, NULL);
    if(strstr(buffer,"Intel") != NULL) { platform = platform_id; break; }
  }
  
  // Get a CPU device on the platform
  cl_device_id device;
  errors = clGetDeviceIDs(platform_ids[platform],CL_DEVICE_TYPE_CPU,1,&device,NULL); error_check(errors);
  errors = clGetDeviceInfo(device,CL_DEVICE_NAME,sizeof(buffer),buffer,NULL); error_check(errors);
  
  // Create a context and a command queue
  cl_context_properties prop[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_ids[platform], 0 };
  printf("Creating an OpenCL context...\n"); fflush(stdout);
  cl_context context = clCreateContext(prop,1,&device,&pfn_notify,NULL,&errors); error_check(errors);
  //cl_context context = clCreateContextFromType(prop,CL_DEVICE_TYPE_CPU,&pfn_notify,NULL,&errors); error_check(errors);
  printf("...done!\n"); fflush(stdout);
  cl_command_queue queue = clCreateCommandQueue(context,device,CL_QUEUE_PROFILING_ENABLE,&errors); error_check(errors);
  
  // Load and compile the kernel
  char *source = get_source("saxpy_device.cl");
  cl_program program = clCreateProgramWithSource(context,1,(const char **)&source,NULL,&errors); error_check(errors);
  errors = clBuildProgram(program,0,NULL,"-cl-single-precision-constant",NULL,NULL);
  
  // Get and print the compiler log
  char* log;
  size_t log_size;
  clGetProgramBuildInfo(program,device,CL_PROGRAM_BUILD_LOG,0,NULL,&log_size);
  log = (char*)malloc((log_size+1)*sizeof(char));
  clGetProgramBuildInfo(program,device,CL_PROGRAM_BUILD_LOG,log_size,log, NULL);
  log[log_size] = '\0';
  if (strcmp(log,"\n") != 0 && strcmp(log,"") != 0) { printf("--------- \n--- Compilation log:\n--------- \n%s\n",log); }
  free(log);
  error_check(errors);
    
  // Create a device pointer for y (zero-copy)
  cl_mem device_y = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR, (2048*2048)*sizeof(float), y, &errors); error_check(errors);
  //cl_mem device_y = clCreateBuffer(context, CL_MEM_READ_WRITE, (2048*2048)*sizeof(float), NULL, &errors); error_check(errors);
  
  // Create a device pointer for x (zero-copy)
  cl_mem device_x = clCreateBuffer(context, CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR, (2048*2048)*sizeof(float), x, &errors); error_check(errors);
  //cl_mem device_x = clCreateBuffer(context, CL_MEM_READ_WRITE, (2048*2048)*sizeof(float), NULL, &errors); error_check(errors);
  
  //errors = clEnqueueWriteBuffer(queue, device_y, CL_TRUE, 0, (2048*2048)*sizeof(float), y, 0, NULL, NULL); error_check(errors);
  //clFinish(queue);  
  //errors = clEnqueueWriteBuffer(queue, device_x, CL_TRUE, 0, (2048*2048)*sizeof(float), x, 0, NULL, NULL); error_check(errors);
  //clFinish(queue);  
  // Create the kernel
  cl_kernel kernel_saxpy_43_0 = clCreateKernel(program, "kernel_saxpy_43_0", &errors); error_check(errors);
  
  // Set all the arguments to the kernel function
  int num_args = 0;
  clSetKernelArg(kernel_saxpy_43_0,num_args+0,sizeof(device_y),(void*)&device_y);
  clSetKernelArg(kernel_saxpy_43_0,num_args+1,sizeof(device_x),(void*)&device_x);
  clSetKernelArg(kernel_saxpy_43_0,num_args+2,sizeof(a),(void*)&a);
  
  // Start the kernel
  size_t global_worksize[] = {4194304};
  errors = clEnqueueNDRangeKernel(queue,kernel_saxpy_43_0,1,NULL,global_worksize,NULL,0,NULL,&event); error_check(errors);
  
  // Synchronize and clean-up the kernel
  clFinish(queue);
  clReleaseKernel(kernel_saxpy_43_0);
  
  // Perform a zero-copy of y from device to host
  void* pointer_to_y = clEnqueueMapBuffer(queue,device_y,CL_TRUE,CL_MAP_READ,(0),(4194304)*sizeof(float),0,NULL,NULL,&errors); error_check(errors);
  clEnqueueUnmapMemObject(queue,device_y,pointer_to_y,0,NULL,NULL);
  clFinish(queue);
  
  // Clean up GPU arrays
  clReleaseMemObject(device_y);
  
  // Clean up GPU arrays
  clReleaseMemObject(device_x);
  
  // Clean-up the OpenCL context
	clReleaseCommandQueue(queue);
  clReleaseProgram(program);
  clReleaseContext(context);
  fflush(stdout);

}
